from typing import List, Dict, Optional, Tuple
import enum
from dataclasses import dataclass, field
from dataclasses_json import config, DataClassJsonMixin
from collections import defaultdict
import logging
from datetime import datetime, timedelta

PORT = 3356

@dataclass(frozen=True)
class Team(DataClassJsonMixin):
    name: str
    hostname: str

@dataclass(frozen=True)
class Service(DataClassJsonMixin):
    name: str
    port: int

@dataclass(frozen=True)
class Target(DataClassJsonMixin):
    team: Team
    service: Service
    flag_id: Optional[str] = None
    tick: Optional[int] = None

    def same_process(self, other: 'Target') -> bool:
        return self.team == other.team and self.service == other.service

class SubmissionResult(enum.Enum):
    OK = 'OK'
    INVALID = 'INVALID'
    UNKNOWN = 'UNKNOWN'
    TOO_OLD = 'TOO_OLD'
    SELF = 'SELF'
    ALREADY_SUBMITTED = 'ALREADY_SUBMITTED'

class ScriptResult(enum.Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    TIMEOUT = "TIMEOUT"
    #ONGOING = "ONGOING"

@dataclass(frozen=True)
class ScriptMetrics(DataClassJsonMixin):
    script: str
    script_hash: str
    corpus_hash: Optional[str]
    host: str
    target: Target
    start_time: datetime
    duration: timedelta = field(metadata=config(encoder=lambda t: t.total_seconds(), decoder=lambda f: timedelta(seconds=f)))
    status: ScriptResult
    exit_code: int
    flags_seen: int
    flags: Tuple[str, ...]

@dataclass(frozen=True)
class Submission(DataClassJsonMixin):
    flag: bytes = field(metadata=config(encoder=lambda b: b.decode('latin-1'), decoder=lambda s: s.encode('latin-1')))
    target: Target
    script: str

@dataclass
class SubmissionLog:
    submission: Submission
    result: SubmissionResult

@dataclass
class ScriptProgress(DataClassJsonMixin):
    runs: int = 0

@dataclass
class TargetStatus(DataClassJsonMixin):
    tick_first_seen: int
    tick_last_seen: int
    script_status: Dict[str, ScriptProgress] = field(default_factory=lambda: defaultdict(ScriptProgress))
    retired: bool = False

@dataclass
class GameStatus:
    tick: int
    targets: List[Target]

@dataclass
class ScriptStatus(DataClassJsonMixin):
    filename: str
    service_name: str
    tick_first_seen: int
    tick_last_seen: int

@dataclass
class ShooterStatus(DataClassJsonMixin):
    tick: int
    targets: Dict[Target, TargetStatus] = field(metadata=config(
        encoder=lambda d: [{"target": t.to_dict(), "status": s.to_dict()} for t, s in d.items()],
        decoder=lambda l: {Target.from_dict(e['target']): TargetStatus.from_dict(e['status']) for e in l},
    ))
    script_info: Dict[str, ScriptStatus]  # keyed on hash
    tick_timeout: int
    retry_timeout: int
    error_log: str

def setup_logging() -> None:
    """installs logging configuration. run this before any other code"""
    root = logging.getLogger()
    root.setLevel(logging.INFO)
    logging.getLogger('quickscope').setLevel('DEBUG')
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    fm = logging.Formatter(
        "%(asctime)s - %(name)-25s - %(funcName)-10s - %(levelname)-5s"
        + " - %(message)s")
    ch.setFormatter(fm)
    root.addHandler(ch)



from typing import Optional, List
import subprocess
import nclib
import time
import unittest
import quickscope
from quickscope.common import *
import random
import threading
import concurrent.futures
from pathlib import Path

BASE_DIR = Path(__file__).absolute().parent

kind_str = {
        SubmissionResult.OK: "O",
        SubmissionResult.INVALID: "I",
        SubmissionResult.UNKNOWN: "U",
        SubmissionResult.TOO_OLD: "T",
        SubmissionResult.SELF: "S",
        SubmissionResult.ALREADY_SUBMITTED: "A",
}

str_kind = {b: a for a, b in kind_str.items()}

def gen_flag(kind: SubmissionResult, num: Optional[int] = None) -> bytes:
    if num is None:
        num = random.randrange(10**10)
    result = f'FLAG{kind_str[kind]}{num:010d}'.encode()
    assert len(result) == 15
    return result

class MyTracker(quickscope.tracker.Tracker):
    def __init__(self):
        super().__init__()
        self.next_game_status = GameStatus(1, [])
        self.log_by_flag = {}

    FLAG_REGEX = b'FLAG[OIUTSA][0-9]{10}'

    def submit_flags(self, flags):
        result = []
        for submission in flags:
            status = str_kind[chr(submission.flag[4])]
            log = SubmissionLog(submission, status)
            result.append(log)
            self.log_by_flag[submission.flag] = log
        return result

    def get_status(self):
        return self.next_game_status

    def instrument_targets(self, targets):
        return targets

    def perform_tick(self, targets: List[Target]):
        self.next_game_status = GameStatus(self.next_game_status.tick + 1, targets)
        self.ingest_status(self.get_status())

    def run_until(self, port, predicate, timeout):
        #scraper_thread = threading.Thread(target=self.scraper_thread)
        #scraper_thread.start()

        server = nclib.server.TCPServer((self.BIND_TO, port))
        timed_out = False

        def wait():
            nonlocal timed_out
            for _ in range(10):
                time.sleep(timeout / 10)
                if predicate():
                    server.close()
                    return
            timed_out = True
            server.close()
        threading.Thread(target=wait).start()

        with concurrent.futures.ThreadPoolExecutor(self.WORKER_THREADS) as executor:
            for client in server:
                executor.submit(self.handle, client)

        if timed_out:
            raise TimeoutError("Failed to reach predicate")

class Tests(unittest.TestCase):
    def test_basic(self):
        tracker = MyTracker()

        team = Team("test_team", "hostname")
        service = Service("test_svc", 1234)
        flag = gen_flag(SubmissionResult.OK)
        tracker.perform_tick([Target(team, service, flag.decode())])
        port = random.randrange(0x2000, 0x8000)
        with subprocess.Popen(['quickscope', '--server', f'localhost:{port}', '--script', BASE_DIR / 'script_basic.py', '--everyone']):
            tracker.run_until(port, lambda: flag in tracker.log_by_flag, 4)
            assert tracker.log_by_flag[flag].result == SubmissionResult.OK




import quickscope
import unittest
import os

basedir = os.path.dirname(os.path.abspath(__file__))

class TestIgnores(unittest.TestCase):
    def test_ignores(self):
        manager = quickscope.shooter.CorpusManager(os.path.join(basedir, 'corpora', 'ignores'), '', 1, quickscope.shooter.Everyone())
        scripts = set(child.script_name for child in manager.children)
        assert scripts == {os.path.join(basedir, 'corpora', 'ignores', stem) for stem in ('a.py', 'c.b', 'subdir1/a.py')}



from typing import Optional, List, Dict, Set, Any
from collections import defaultdict
import random
import concurrent.futures
import threading
import nclib
import io
import time
import pickle
import os
import argparse
import logging
import queue
import sys
import signal
import IPython

from .common import *


logger = logging.getLogger(__name__)

class ArgsNamespace(argparse.Namespace):
    database: Optional[str]
    flag_log: Optional[str]
    shell: bool

class Tracker:
    # HERE'S WHAT YOU IMPLEMENT
    FLAG_REGEX: bytes = NotImplemented

    def get_status(self) -> GameStatus:
        raise NotImplementedError

    def submit_flags(self, flags: List[Submission]) -> List[SubmissionLog]:
        raise NotImplementedError

    def instrument_targets(self, targets: List[Target]) -> List[Target]:
        """
        This is called just before shooting at the given target list - each target in the list will be shot at once.

        This gives the tracker the ability to record that shooting at a target is occurring and to modify the target
        list, for example, to remove targets with rate-limiting imposed.
        """
        raise NotImplementedError

    TICK_TIMEOUT = 3
    RETRY_TIMEOUT = 3
    BIND_TO = '0.0.0.0'
    WORKER_THREADS = 8

    # DON'T TOUCH THIS

    @classmethod
    def arg_parser(cls) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        parser.add_argument('--database', help='Use this file for state persistence')
        parser.add_argument('--flag-log', help='Use this file to log all seen flags')
        parser.add_argument('--shell', help='Drop into an interactive shell instead of running the tracker', action='store_true')
        return parser

    @classmethod
    def main(cls) -> None:
        args = cls.arg_parser().parse_args(namespace=ArgsNamespace())

        setup_logging()

        if args.database is not None and os.path.exists(args.database):
            with open(args.database, 'rb') as fp:
                inst = pickle.load(fp)
        else:
            inst = cls()

        if args.shell:
            IPython.start_ipython(argv=[], user_ns=locals() | globals())  # type: ignore
        else:
            try:
                inst.boot()
                inst.run(args)
            except KeyboardInterrupt:
                pass
            finally:
                if args.database is not None:
                    with open(args.database, 'wb') as fp:
                        pickle.dump(inst, fp)
                    print(f'Saved database to {args.database}')

    def __init__(self) -> None:
        self.lock: threading.Lock = threading.Lock()
        self.tick: int = -1
        self.targets: Dict[Target, TargetStatus] = {}
        self.script_info: Dict[str, ScriptStatus] = {}  # keyed on hash
        self.script_queues: Dict[str, List[Target]] = defaultdict(list)
        self.submit_buffer: List[Submission] = []
        self.submit_buffer_lock: threading.Lock = threading.Lock()
        self.flag_log: Optional[io.IOBase] = None
        self.script_metrics_queue: queue.Queue[ScriptMetrics] = queue.Queue(maxsize=300)
        self.script_metrics_subscribers: Set[nclib.Netcat] = set()

        self.submit_thread: threading.Thread = threading.Thread(target=self._submit_thread, daemon=True)
        self.metrics_thread: threading.Thread = threading.Thread(target=self._metrics_thread, daemon=True)
        self.scraper_thread: threading.Thread = threading.Thread(target=self._scraper_thread, daemon=True)

        root = logging.getLogger()
        buf = io.StringIO()
        sh = logging.StreamHandler(buf)
        fm = logging.Formatter(
            "%(asctime)s - %(name)-25s - %(funcName)-10s - %(levelname)-5s - %(message)s")
        sh.setFormatter(fm)
        sh.setLevel('WARNING')
        root.addHandler(sh)

        self.logging_memory: io.StringIO = buf

        assert b'\n' not in self.FLAG_REGEX

    def boot(self) -> None:
        logger.info("Tracker starting on %s:%s", self.BIND_TO, PORT)
        self.submit_thread.start()
        self.metrics_thread.start()
        self.scraper_thread.start()

        def bye(*args: Any) -> None:
            sys.exit()

        signal.signal(signal.SIGTERM, bye)

    def __getstate__(self) -> Dict[str, Any]:
        return {
            '__version__': 1,
            'tick': self.tick,
            'targets': self.targets,
            'script_info': self.script_info,
            'script_queues': self.script_queues,
            'submit_buffer': self.submit_buffer,
        }

    def __setstate__(self, state: Dict[str, Any]) -> None:
        version = state.pop('__version__')
        if version == 1:
            self.__init__() # type: ignore
            self.tick = state['tick']
            self.targets = state['targets']
            self.script_info = state['script_info']
            self.script_queues = state['script_queues']
            self.submit_buffer = state['submit_buffer']
        else:
            raise NotImplementedError("Unsupported database")

    def run(self, args: ArgsNamespace) -> None:
        server = nclib.server.TCPServer((self.BIND_TO, PORT))
        if args.flag_log is not None:
            self.flag_log = open(args.flag_log, 'ab')
        try:
            with concurrent.futures.ThreadPoolExecutor(self.WORKER_THREADS) as executor:
                for client in server:
                    executor.submit(self.handle, client)
        finally:
            if self.flag_log is not None:
                self.flag_log.close()
                self.flag_log = None

    def handle(self, sock: nclib.Netcat) -> None:
        line = b''
        try:
            line = sock.readln(max_size=100, timeout=1).strip()
            if line == b'submit':
                self.serve_submit(sock)
            elif line == b'getregex':
                self.serve_getregex(sock)
            elif line == b'gettargets':
                self.serve_gettargets(sock)
            elif line == b'gettargetsdumb':
                self.serve_gettargetsdumb(sock)
            elif line == b'getstatus':
                self.serve_getstatus(sock)
            elif line == b'pubmetrics':
                self.serve_pubmetrics(sock)
            elif line == b'submetrics':
                self.script_metrics_subscribers.add(sock)
            else:
                raise Exception("not a command", line)
        except Exception:
            logger.exception("Exception in handle()")
        finally:
            if line != b'submetrics':
                sock.close()

    def serve_pubmetrics(self, sock: nclib.Netcat) -> None:
        lines = sock.recvall(timeout=1)
        if not lines:
            return
        for line in lines.splitlines():
            try:
                metric = ScriptMetrics.from_json(line.decode())
                self.script_metrics_queue.put(metric)
            except Exception:
                logger.exception("Exception during metrics parsing")

    def _metrics_thread(self) -> None:
        while True:
            metric = self.script_metrics_queue.get()
            try:
                for sock in list(self.script_metrics_subscribers):
                    try:
                        sock.sendline(metric.to_json().encode())
                    except (nclib.NetcatError, OSError):
                        self.script_metrics_subscribers.remove(sock)
            except Exception:
                logger.exception("Exception during metrics broadcast")
                time.sleep(10)

    def serve_submit(self, sock: nclib.Netcat) -> None:
        lines = sock.recvall(timeout=1)
        if not lines:
            return
        submissions = []
        for line in lines.splitlines():
            try:
                submission = Submission.from_json(line.decode())
                submissions.append(submission)
                if self.flag_log is not None:
                    self.flag_log.write(submission.flag + b'\n')

            except Exception:
                logger.exception("Exception during submit parsing")

        if self.flag_log is not None:
            self.flag_log.flush()

        try:
            results = self.submit_flags(submissions)
        except Exception:
            logger.exception("Exception during submit")
            results = []
        self.process_successful_submissions(submissions, results, sock)

    def _submit_thread(self) -> None:
        while True:
            time.sleep(10)
            next_batch = None
            with self.submit_buffer_lock:
                if len(self.submit_buffer) > 0:
                    next_batch = self.submit_buffer
                    self.submit_buffer = []
            if next_batch is not None:
                try:
                    results = self.submit_flags(next_batch)
                except Exception:
                    logger.exception("Exception during submit retry")
                    results = []
                self.process_successful_submissions(next_batch, results, None)

    def process_successful_submissions(
            self,
            submissions: List[Submission],
            results: List[SubmissionLog],
            sock: Optional[nclib.Netcat]=None
    ) -> None:
        submissions_set = set(submissions)
        for result in results:
            try:
                submissions_set.remove(result.submission)
            except KeyError:
                pass
            if result.result == SubmissionResult.SELF:
                self.untarget_host_port(result.submission.target)
            elif result.result == SubmissionResult.OK:
                logger.info(
                    "Got points on %s:%s",
                    result.submission.target.team.name,
                    result.submission.target.service.name
                )
                if sock is not None:
                    try:
                        sock.sendline(result.submission.flag)
                    except BrokenPipeError:
                        pass
                self.untarget_target(result.submission.target)
            elif result.result == SubmissionResult.ALREADY_SUBMITTED:
                self.untarget_target(result.submission.target)
            elif result.result == SubmissionResult.INVALID:
                logger.warning("Getting bogus flags for %s", result.submission.target)
                pass
            elif result.result == SubmissionResult.UNKNOWN:
                pass
            elif result.result == SubmissionResult.TOO_OLD:
                pass
            else:
                logger.error("Bad submission result: %s (is your submitter misbehaving?)",
                             result.result)

        with self.submit_buffer_lock:
            self.submit_buffer.extend(submissions_set)

    def serve_getregex(self, sock: nclib.Netcat) -> None:
        sock.sendln(self.FLAG_REGEX)

    def serve_gettargets(self, sock: nclib.Netcat) -> None:
        script_id = sock.readln(max_size=100, timeout=1).strip().decode()
        script_name = sock.readln(max_size=100, timeout=1).strip().decode()
        service_name = sock.readln(max_size=100, timeout=1).strip().decode()
        n = int(sock.readln(max_size=100, timeout=1).strip().decode())
        if script_id not in self.script_info:
            with self.lock:
                if script_id not in self.script_info:
                    self.script_info[script_id] = ScriptStatus(
                        filename=script_name,
                        service_name=service_name,
                        tick_first_seen=self.tick,
                        tick_last_seen=self.tick,
                    )
        self.script_info[script_id].tick_last_seen = self.tick
        targets = self.get_targets_for_script(script_id, service_name, n)
        targets = self.instrument_targets(targets)
        for target in targets:
            sock.sendln(target.to_json().encode())

    def serve_gettargetsdumb(self, sock: nclib.Netcat) -> None:
        service_name = sock.readln(max_size=100, timeout=1).strip().decode()

        targets = self.get_targets_for_tick(service_name, self.tick)
        targets = self.instrument_targets(targets)
        for target in targets:
            sock.sendln(target.to_json().encode())

    def serve_getstatus(self, sock: nclib.Netcat) -> None:
        result = ShooterStatus(
            tick=self.tick,
            script_info=self.script_info,
            targets=dict(self.targets),
            tick_timeout=self.TICK_TIMEOUT,
            retry_timeout=self.RETRY_TIMEOUT,
            error_log=self.logging_memory.getvalue(),
        )
        sock.send(result.to_json().encode())

    def _scraper_thread(self) -> None:
        while True:
            try:
                status = self.get_status()
            except Exception:
                logger.exception("Error getting status")
                time.sleep(10)
                continue

            if status.tick == self.tick:
                time.sleep(10)
                continue
            self.ingest_status(status)

    def ingest_status(self, gamestatus: GameStatus) -> None:
        with self.lock:
            self.tick = gamestatus.tick

            for target in gamestatus.targets:
                if target in self.targets:
                    self.targets[target].tick_last_seen = gamestatus.tick
                else:
                    self.targets[target] = TargetStatus(tick_first_seen=gamestatus.tick, tick_last_seen=gamestatus.tick)

            to_remove = []
            for target, status in self.targets.items():
                if status.tick_last_seen < gamestatus.tick - self.TICK_TIMEOUT:
                    to_remove.append(target)
            for target in to_remove:
                self.targets.pop(target)

    def untarget_host_port(self, bad_target: Target) -> None:
        with self.lock:
            for target, status in self.targets.items():
                if target.same_process(bad_target):
                    status.retired = True

    def untarget_target(self, target: Target) -> None:
        try:
            self.targets[target].retired = True
        except KeyError:
            pass

    def get_targets_for_script(self, script_id: str, service_name: str, n: int) -> List[Target]:
        result = None
        found_one = False
        with self.lock:
            if len(self.script_queues[script_id]) >= n:
                result, self.script_queues[script_id] = self.script_queues[script_id][:n], self.script_queues[script_id][n:]
                found_one = True

        if result is None:
            newstuff = []
            for target, status in self.targets.items():
                if target.service.name != service_name:
                    continue
                found_one = True
                if status.retired:
                    continue
                if status.script_status[script_id].runs >= self.RETRY_TIMEOUT:
                    continue
                status.script_status[script_id].runs += 1
                newstuff.append(target)
            random.shuffle(newstuff)

            with self.lock:
                self.script_queues[script_id].extend(newstuff)
                result, self.script_queues[script_id] = self.script_queues[script_id][:n], self.script_queues[script_id][n:]

        if not found_one:
            logger.warning("Someone requested unknown service %s", repr(service_name))
        return result

    def get_targets_for_tick(self, service_name: str, tick: int) -> List[Target]:
        result = []
        found_one = False
        for target, status in self.targets.items():
            if target.service.name != service_name:
                continue
            found_one = True
            if status.tick_last_seen != tick:
                continue
            result.append(target)

        if not found_one:
            logger.warning("Someone requested unknown service %s", repr(service_name))
        return result



from typing import Tuple, Dict, List, Type
import json.decoder
import html
from collections import defaultdict
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
import logging
import math
import nclib
import http.server


logger = logging.getLogger(__name__)

STATUS_HTML = """<!DOCTYPE html>
<html lang="en-US">
<head>
    <title>Shooter Status</title>
    <script>
        var current_tick = null;
        var current_service = 1;
        function showTickService(tick, service) {
            current_tick = tick;
            current_service = service;

            var cards = document.getElementsByClassName("card");
            for (var i = 0; i < cards.length; i++) {
                var card = cards[i];
                var card_tick = parseInt(card.dataset.tick);
                var card_service = card.dataset.service;
                if (card.dataset.role === "get") {
                    if (card_tick === tick && card_service === service) {
                        card.style.display = '';
                    } else {
                        card.style.display = 'none';
                    }
                } else if (card.dataset.role === "set-service") {
                    if (card_tick === tick) {
                        card.style.display = '';
                    } else {
                        card.style.display = 'none';
                    }

                    if (card_service === service) {
                        card.classList.add("selected")
                    } else {
                        card.classList.remove("selected")
                    }
                } else if (card.dataset.role === "set-tick") {
                    if (card_tick === tick) {
                        card.classList.add("selected")
                    } else {
                        card.classList.remove("selected")
                    }
                }
            }
        }

        window.onload = function () {
            var cards = document.getElementsByClassName("card");
            var last = null;
            var second_last = null;
            for (var i = 0; i < cards.length; i++) {
                var card = cards[i];
                if (card.dataset.role === "set-tick") {
                    let card_tick = parseInt(card.dataset.tick);
                    second_last = last;
                    last = card_tick;
                    card.onclick = function () {
                        console.log("tick", card_tick);
                        showTickService(card_tick, "ALL");
                    };
                } else if (card.dataset.role == "set-service") {
                    let card_service = card.dataset.service;
                    card.onclick = function () {
                        console.log("service", card_service);
                        if (current_service == card_service) {
                            showTickService(current_tick, "ALL");
                        } else {
                            showTickService(current_tick, card_service);
                        }
                    };
                }
            }
            if (second_last !== null) {
                showTickService(second_last, "ALL");
            } else if (last !== null) {
                showTickService(last, "ALL");
            }
        };
    </script>
    <style>
        html {
            background-color: #222222;
            color: white;
        }
        
        #title {
            font-size: 40px;
        }
        
        .card {
            display: inline-block;
            text-align: center;
            background-color: #222228;
            border-width: 1px;
            border-color: #666;
            border-style: solid;
            margin: 5px;
            padding-top: 10px;
        }
        
        .card.selected {
            border-color: orange;
        }
        
        .card-title {
            font-weight: bold;
        }
        
        .card.selectable:hover {
            background-color: #333336;
            cursor: pointer;
        }
    </style>
</head>
<body>
    <div id="title">Shooter Status</div>
    <div id="progress">
        %(progress)s
    </div>
    <div id="services">
        %(services)s
    </div>
    <div id="teams">
        %(teams)s
    </div>
    <div id="title">Error Log</div>
    <div id="errorlog">
        <pre><code>%(errorlog)s</code></pre>
    </div>
</body>
</html>
"""

CARD_HTML = """
<div class="card" data-tick="%(tick)s" data-service="%(service)s" data-role="%(role)s">
    <div class="card-title">%(title)s</div>
    %(chart)s
</div>
"""

from .common import ShooterStatus, ScriptStatus

@dataclass_json
@dataclass
class FlagsReport:
    total_flags: int = 0
    got_flags: int = 0

@dataclass_json
@dataclass
class TickReport:
    expected_shots: int = 0
    total_shots: int = 0
    flags_by_service: Dict[str, FlagsReport] = field(default_factory=lambda: defaultdict(FlagsReport))
    flags_by_team: Dict[str, FlagsReport] = field(default_factory=lambda: defaultdict(FlagsReport))
    flags_by_team_service: Dict[Tuple[str, str], FlagsReport] = field(default_factory=lambda: defaultdict(FlagsReport))

class StatusServer(http.server.ThreadingHTTPServer):
    def __init__(self, bind_address: Tuple[str, int], handler_class: Type["StatusHandler"], qs_server: str):
        super().__init__(bind_address, handler_class)
        self.qs_server: str = qs_server

class StatusHandler(http.server.BaseHTTPRequestHandler):
    server: StatusServer

    def do_GET(self) -> None:
        data = statuspage(self.server.qs_server).encode()
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", 'text/html')
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.wfile.write(data)

def statuspage(server: str) -> str:
    sock = nclib.Netcat(server)
    sock.sendln(b'getstatus')
    recved = sock.recvall()
    try:
        dicts = json.loads(recved.decode())
        data: ShooterStatus = ShooterStatus.from_dict(dicts)
    except json.decoder.JSONDecodeError:
        logger.exception('JSON decode error for status: %r', recved)
        return 'JSON decode error for status: %r' % recved

    sock.close()

    scripts_by_service: defaultdict[str, List[Tuple[str, ScriptStatus]]] = defaultdict(list)
    for script_hash, script_info in data.script_info.items():
        scripts_by_service[script_info.service_name].append((script_hash, script_info))

    ticks: Dict[int, TickReport] = defaultdict(TickReport)
    for target, status in data.targets.items():
        report = ticks[status.tick_first_seen]
        # how many shots can we expect to be made because of this target?
        # how many scripts for this service are active during this tick?
        for script_hash, script_info in scripts_by_service[target.service.name]:
            if status.tick_first_seen <= script_info.tick_first_seen <= status.tick_last_seen or \
                    script_info.tick_first_seen <= status.tick_first_seen <= script_info.tick_last_seen:
                report.expected_shots += data.retry_timeout
                if status.retired:
                    report.total_shots += data.retry_timeout
                else:
                    try:
                        report.total_shots += status.script_status[script_hash].runs
                    except KeyError:
                        pass

        report.flags_by_team[target.team.name].total_flags += 1
        report.flags_by_team_service[(target.team.name, target.service.name)].total_flags += 1
        report.flags_by_service[target.service.name].total_flags += 1
        if status.retired:
            report.flags_by_team[target.team.name].got_flags += 1
            report.flags_by_team_service[(target.team.name, target.service.name)].got_flags += 1
            report.flags_by_service[target.service.name].got_flags += 1

    ticks_html = '\n'.join(CARD_HTML % dict(
        title=f'Tick {tick}',
        chart=build_pie([max(0, ticks[tick].expected_shots - ticks[tick].total_shots), ticks[tick].total_shots], ['yellow', 'blue']),
        tick=tick,
        service=0,
        role="set-tick",
    ) for tick in sorted(ticks))
    services_html = '\n'.join(CARD_HTML % dict(
        title=f'{html.escape(service)}',
        chart=build_pie([max(0, report.total_flags - report.got_flags), report.got_flags], ['red', 'green']),
        tick=tick,
        service=service,
        role="set-service",
    ) for tick, tick_data in ticks.items() for service, report in tick_data.flags_by_service.items()) + '\n'
    teams_html = '\n'.join(CARD_HTML % dict(
        title=f'{html.escape(team)} - {report.got_flags}/{report.total_flags}',
        chart=build_pie([max(0, report.total_flags - report.got_flags), report.got_flags], ['red', 'green']),
        tick=tick,
        service=service,
        role="get",
    ) for tick, tick_data in ticks.items() for (team, service), report in tick_data.flags_by_team_service.items()) + '\n'
    teams_html += '\n'.join(CARD_HTML % dict(
        title=f'{html.escape(team)} - {report.got_flags}/{report.total_flags}',
        chart=build_pie([max(0, report.total_flags - report.got_flags), report.got_flags], ['red', 'green']),
        tick=tick,
        service="ALL",
        role="get",
    ) for tick, tick_data in ticks.items() for team, report in tick_data.flags_by_team.items()) + '\n'

    return STATUS_HTML % dict(progress=ticks_html, services=services_html, teams=teams_html,
                                errorlog=data.error_log)

def endpoint(pAngleInRadians: float, pRadius: float, pCentreOffsetX: float, pCentreOffsetY: float) -> Tuple[float, float]:
    """
    Calculate position of point on circle given an angle, a radius, and the location of the center of the circle
    Zero line points west.
    """
    lCosAngle = math.cos(pAngleInRadians)
    lSinAngle = math.sin(pAngleInRadians)
    lStartLineDestinationX =  pCentreOffsetX - (pRadius * lCosAngle)  # originally said lRadius
    lStartLineDestinationY =  pCentreOffsetY - (pRadius * lSinAngle)

    return (lStartLineDestinationX, lStartLineDestinationY)

# https://drumcoder.co.uk/blog/2010/nov/16/python-code-generate-svg-pie-chart/
def build_pie(slices: List[int], colors: List[str]) -> str:
    if sum(slices) == 0:
        slices = [1]
        colors = ['white']

    OFFSET_X = 75
    OFFSET_Y = 75
    RADIUS = 65
    INNER_RADIUS = 30
    DEGREES_IN_CIRCLE = 360.0

    current_angle = 0.
    total = sum(slices)
    path_elem = ""

    for gradient, slice in zip(colors, slices):
        degrees = (DEGREES_IN_CIRCLE / total) * slice

        if degrees <= 0:
            pass
        else:
            route = 0
            if degrees > 180:
                route = 1
            if degrees >= 360:
                degrees = 359.99

            start_inner_x, start_inner_y = endpoint(current_angle, INNER_RADIUS, OFFSET_X, OFFSET_Y)
            start_outer_x, start_outer_y = endpoint(current_angle, RADIUS, OFFSET_X, OFFSET_Y)
            radians = math.radians(degrees)
            current_angle += radians
            end_inner_x, end_inner_y = endpoint(current_angle, INNER_RADIUS, OFFSET_X, OFFSET_Y)
            end_outer_x, end_outer_y = endpoint(current_angle, RADIUS, OFFSET_X, OFFSET_Y)

            line_1 = "M%f,%f L%f,%f" % (start_inner_x, start_inner_y, start_outer_x, start_outer_y)
            arc_1 = "A%d,%d 0 %d,1 %f %f" % (RADIUS, RADIUS, route, end_outer_x, end_outer_y)
            line_2 = "L%f,%f" % (end_inner_x, end_inner_y)
            arc_2 = "A%d,%d 0 %d,0 %f %f" % (INNER_RADIUS, INNER_RADIUS, route, start_inner_x, start_inner_y)

            path = "%s %s %s %s Z" % (line_1, arc_1, line_2, arc_2)
            path_elem += "<path d='%s' style='fill:url(#%s);'/>" % (path, gradient)

    svg = """
    <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="150" height="150">
      <defs>
        <radialGradient id="green" r="65%%" cx="0" cy="0" spreadMethod="pad" gradientUnits="userSpaceOnUse">
          <stop offset="0%%"   stop-color="#00ee00" stop-opacity="1"/>
          <stop offset="100%%" stop-color="#006600" stop-opacity="1" />
        </radialGradient>
        <radialGradient id="red" r="65%%" cx="0" cy="0" spreadMethod="pad" gradientUnits="userSpaceOnUse">
          <stop offset="0%%"   stop-color="#ff0000" stop-opacity="1"/>
          <stop offset="100%%" stop-color="#880000" stop-opacity="1" />
        </radialGradient>
        <radialGradient id="yellow" r="65%%" cx="0" cy="0" spreadMethod="pad" gradientUnits="userSpaceOnUse">
          <stop offset="0%%"   stop-color="#ffcc00" stop-opacity="1"/>
          <stop offset="100%%" stop-color="#886600" stop-opacity="1" />
        </radialGradient>
        <radialGradient id="blue" r="65%%" cx="0" cy="0" spreadMethod="pad" gradientUnits="userSpaceOnUse">
          <stop offset="0%%"   stop-color="#55aaff" stop-opacity="1"/>
          <stop offset="100%%" stop-color="#0022cc" stop-opacity="1" />
        </radialGradient>
        <radialGradient id="blue" r="65%%" cx="0" cy="0" spreadMethod="pad" gradientUnits="userSpaceOnUse">
          <stop offset="0%%"   stop-color="#ffffff" stop-opacity="1"/>
          <stop offset="100%%" stop-color="#aaaaaa" stop-opacity="1" />
        </radialGradient>
      </defs>

      %s
    </svg>
    """ % (path_elem,)

    return svg



from typing import List, Optional, Tuple, Union, Set, Dict, Iterable, Any, Protocol
import subprocess
import pathlib
import datetime
import select
import time
import os
import sys
import argparse
import re
import hashlib
import threading
import psutil
from dataclasses import dataclass
import queue
import resource
import logging
import signal
import platform

import nclib
from readerwriterlock import rwlock

from .common import PORT, Target, Submission, Team, Service, setup_logging, ScriptResult, ScriptMetrics
from .statuspage import statuspage, StatusServer, StatusHandler


logger = logging.getLogger(__name__)

SERVICE_NAME_RE = re.compile(br'x-service-name: ([-\w_=+,./?]+)')

parser = argparse.ArgumentParser('quickscope')
parser.add_argument('--corpus', help='A directory of exploits to fire')
parser.add_argument('--script', help='A single script to fire')
parser.add_argument('--server', help='The shooter tracker to connect to', default='172.31.2.25:%d' % PORT)
parser.add_argument('--procs', help='The level of parallelism to use', type=int)
parser.add_argument('--adaptive-procs', help='Measure system load to determine level of parallelism',
                    action='store_true')
parser.add_argument('--host', help='The single host to shoot at')
parser.add_argument('--port', help='The single port to shoot at', type=int)
parser.add_argument('--flag-id', help='The single flag id to shoot at')
parser.add_argument('--everyone', help='Fire at all live targets and quit', action='store_true')
parser.add_argument('--forever', help='Fire at all live targets until the end of time', action='store_true')
parser.add_argument('--batch', help='Tunes the number of targets which are claimed at once', type=int, default=1)
parser.add_argument('--logdir', help='Directory to store logs in')
parser.add_argument('--no-stdout', help='Disable printing exploit logs to stdout', action='store_true')
parser.add_argument('--timeout', help='Timeout (seconds) for each exploit run', type=int)
parser.add_argument('--mem-limit', help='Memory limit to impose on exploit scripts (GB)', type=float, default=2)
parser.add_argument('--status-html', help='Dump status to file as html and exit')
parser.add_argument('--status-server', help='Serve status on given addr:port forever')
parser.add_argument('--corpus-hints', help='Use the name of the first directory level inside a corpus as a hint for the name of the service a script targets', action='store_true')
parser.add_argument('--submit', help='Pipe stdin to the flag submission service', action='store_true')
parser.add_argument('--sync-command', help='Command to run on an interval while no shots are running in order to perform maintenence')
parser.add_argument('--sync-interval', help='Time in seconds between sync-command runs', type=int, default=60)

class ShooterArgs(argparse.Namespace):
    corpus: Optional[str]
    script: Optional[str]
    server: str
    procs: Optional[int]
    adaptive_procs: bool
    host: Optional[str]
    port: Optional[int]
    flag_id: Optional[str]
    everyone: Optional[bool]
    forever: Optional[bool]
    batch: int
    logdir: Optional[str]
    no_stdout: bool
    timeout: Optional[int]
    mem_limit: float
    status_html: Optional[str]
    status_server: Optional[str]
    corpus_hints: bool
    submit: bool
    service: Optional[str]
    team: Optional[str]
    sync_command: Optional[str]
    sync_interval: int

class NotAnExploit(ValueError):
    pass

@dataclass
class Single:
    host: str
    port: int
    flag_id: Optional[str]

@dataclass
class Everyone:
    pass

@dataclass
class Forever:
    pass

TargetMode = Union[Single, Everyone, Forever]

def parse_target_mode(args: ShooterArgs) -> TargetMode:
    results: List[TargetMode] = []
    if args.forever:
        results.append(Forever())
    if args.everyone:
        results.append(Everyone())
    if args.flag_id is not None or args.host is not None or args.port is not None:
        if args.host is None or args.port is None:
            sys.stderr.write("Must provide both of --host/--port or none at all\n")
            sys.stderr.flush()
            sys.exit(1)
        results.append(Single(host=args.host, port=args.port, flag_id=args.flag_id))

    if len(results) != 1:
        sys.stderr.write("Must provide exactly one of --everyone, --forever, or --host/--port/--flag-id\n")
        sys.stderr.flush()
        sys.exit(1)

    return results[0]

def get_flag_regex(server: str) -> re.Pattern[bytes]:
    while True:
        try:
            sock = nclib.Netcat(server)
        except nclib.NetcatError:
            time.sleep(3)
            continue
        else:
            break
    sock.sendln(b'getregex')
    return re.compile(sock.readln().strip())

class ScriptManager:
    def __init__(self, script: str, server: str, batch: int, target_mode: TargetMode, script_shortname: str, toplevel: bool = False, service_hint: Optional[str] = None):
        self.server: str = server
        self.script_name: str = script
        self.script_shortname = script_shortname
        self.toplevel: bool = toplevel
        service_name, script_hash = get_script_service_name_and_hash(script, service_hint)
        if script_hash is None:
            raise NotAnExploit(f"Please put `x-service-name: service name` somewhere in {script}, or put it in a .shooterignore file")
        if service_name is None:
            raise NotAnExploit()
        self.service_name: str = service_name
        self.script_hash: str = script_hash
        if not os.access(self.script_name, os.X_OK):
            if not self.script_name.endswith('.py'):
                raise NotAnExploit(script + " is not executable - how do I run it?")

        self.batch: int = batch
        self.target_mode: TargetMode = target_mode
        self.eof: bool = False
        self.buffer: List[Target] = []

    def _buffer_targets(self) -> None:
        if self.eof:
            return

        if isinstance(self.target_mode, Forever):
            sock = nclib.Netcat(self.server, raise_timeout=True)
            sock.sendln(b'gettargets')
            sock.sendln(self.script_hash)
            sock.sendln(self.script_name)
            sock.sendln(self.service_name)
            sock.sendln(str(self.batch).encode())

            try:
                lines = sock.recvall(timeout=10).splitlines()
            except nclib.NetcatTimeout:
                logger.warning("Server did not respond with target list")
                self.eof = True
                return

            self.buffer.extend(Target.from_json(line.decode()) for line in lines)
            if not self.buffer:
                self.eof = True
        elif isinstance(self.target_mode, Everyone):
            sock = nclib.Netcat(self.server, raise_timeout=True)
            sock.sendln(b'gettargetsdumb')
            sock.sendln(self.service_name)

            try:
                lines = sock.recvall(timeout=10).splitlines()
            except nclib.NetcatTimeout:
                logger.warning("Server did not respond with target list")
                self.eof = True
                return

            self.buffer.extend(Target.from_json(line.decode()) for line in lines)
            self.eof = True
        elif isinstance(self.target_mode, Single):
            self.buffer.append(Target(
                team=Team(name=self.target_mode.host, hostname=self.target_mode.host),
                service=Service(name=str(self.target_mode.port), port=self.target_mode.port),
                flag_id=self.target_mode.flag_id,
            ))
            self.eof = True

    def __iter__(self) -> "ScriptManager":
        return self

    def __next__(self) -> Tuple[str, str, Target]:
        while True:
            if not self.buffer:
                try:
                    self._buffer_targets()
                except nclib.NetcatError:
                    logger.exception('Failed to retrieve targets')
            if not self.buffer:
                if isinstance(self.target_mode, Forever) and self.toplevel:
                    self.eof = False
                    time.sleep(4)
                    continue
                raise StopIteration()
            return self.script_name, self.script_shortname, self.buffer.pop()

class CorpusManager:
    def __init__(self, corpus: str, server: str, batch: int, target_mode: TargetMode, subdir_hints: bool = False):
        if isinstance(target_mode, Single):
            sys.stderr.write("Error: cannot specify --corpus and --host/--port/--flag-id\n")
            sys.stderr.flush()
            sys.exit(1)

        self.corpus: str = corpus
        self.server: str = server
        self.batch: int = batch
        self.target_mode: TargetMode = target_mode
        self.subdir_hints: bool = subdir_hints

        self.children: List[ScriptManager] = []
        self.eof: bool = False
        self._collect()

    def _collect(self) -> None:
        if self.eof:
            return
        ignores: List[str] = []
        for root, dirs, files in os.walk(self.corpus):
            subroot = pathlib.Path(root).relative_to(self.corpus)
            subdir_hint = subroot.parts[0] if self.subdir_hints and len(subroot.parts) > 0 and subroot.parts[0] != '.' else None

            if '.shooterignore' in files:
                with open(os.path.join(root, '.shooterignore'), 'r') as fp:
                    lines = fp.read().splitlines()
                ignores.extend(os.path.join(root, line) for line in lines)
                ignores.extend(os.path.join(root, '**', line) for line in lines)

            for stem in files:
                if stem == '.shooterignore':
                    continue
                filename = os.path.join(root, stem)
                if any(pathlib.Path(filename).match(pattern) for pattern in ignores):
                    continue
                try:
                    child = ScriptManager(filename, self.server, self.batch, self.target_mode, str(pathlib.Path(filename).relative_to(self.corpus)), service_hint=subdir_hint)
                except NotAnExploit as e:
                    if e.args:
                        logger.warning("%s", e.args[0])
                else:
                    self.children.append(child)
            dirs[:] = [stem for stem in dirs if not any((pathlib.Path(root) / stem).match(pattern) for pattern in ignores)]
        self.eof = True

    def __iter__(self) -> "CorpusManager":
        return self

    def __next__(self) -> Tuple[str, str, Target]:
        while True:
            if not self.children:
                self._collect()
            try:
                child = self.children.pop(0)
            except IndexError:
                if isinstance(self.target_mode, Forever):
                    self.eof = False
                    time.sleep(4)
                    continue
                else:
                    raise StopIteration()

            try:
                result = next(child)
            except StopIteration:
                continue

            self.children.append(child)
            return result

class Pool(Protocol):
    def apply(self, iterator: Iterable[Tuple[str, str, Target]], *args: Any, **kwargs: Any) -> None:
        ...

class SynchronousPool:
    def __init__(self, lock: rwlock.RWLockWrite):
        self.lock = lock

    def apply(self, iterator: Iterable[Tuple[str, str, Target]], *args: Any, **kwargs: Any) -> None:
        for script, script_shortname, target in iterator:
            kwargss: Dict[str, Any] = kwargs | {'script_shortname': script_shortname}
            with self.lock.gen_rlock():
                shoot(script, target, *args, **kwargss)

class AsyncPool:
    def __init__(self, lock: rwlock.RWLockWrite, procs: int):
        self.queue: "queue.Queue[Tuple[str, str, Target]]" = queue.Queue(maxsize=1)
        self.threads = [threading.Thread(target=self.worker, daemon=True) for _ in range(procs)]
        self.args: Tuple[Any, ...] = ()
        self.kwargs: Dict[str, Any] = {}
        self.lock = lock
        for thread in self.threads:
            thread.start()

    def apply(self, iterator: Iterable[Tuple[str, str, Target]], *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs
        for script, shortname, target in iterator:
            self.queue.put((script, shortname, target), block=True)
        self.queue.join()

    def worker(self) -> None:
        while True:
            script, shortname, target = self.queue.get(block=True)
            try:
                kwargs: Dict[str, Any] = self.kwargs | {'script_shortname': shortname}
                with self.lock.gen_rlock():
                    shoot(script, target, *self.args, **kwargs)
            except:
                logger.exception("Failed to shoot")
            finally:
                self.queue.task_done()

class AdaptivePool:
    # warning: instances of this class will never be garbage collected

    def __init__(self, lock: rwlock.RWLockWrite) -> None:
        self.cpu_utilization: float = 0.0
        self.mem_utilization: float = 0.0
        self.watcher_thread: threading.Thread = threading.Thread(target=self.load_watcher, daemon=True)
        self.watcher_thread.start()
        self.worker_threads: List[threading.Thread] = []

        self.args: Tuple[Any, ...] = ()
        self.kwargs: Dict[str, Any] = {}
        self.queue: queue.Queue[Tuple[str, str, Target]] = queue.Queue(maxsize=1)
        self.rlock = lock
        self.lock: threading.Lock = threading.Lock()
        self.live_tasks: int = 0
        self.target_threads: int = 1

    def load_watcher(self) -> None:
        while True:
            self.cpu_utilization = psutil.cpu_percent(1) / 100.0
            self.mem_utilization = psutil.virtual_memory().percent / 100.0

            if self.live_tasks == self.target_threads and self.cpu_utilization < 0.9 and self.mem_utilization < 0.75:
                self.target_threads += 1
                logger.info("Increasing target threads to %s", self.target_threads)
            elif (self.cpu_utilization > 0.99 or self.mem_utilization > 0.95) and self.target_threads > 1:
                self.target_threads -= 1
                logger.info("Decreasing target threads to %s", self.target_threads)

            while self.worker_threads and not self.worker_threads[-1].is_alive():
                self.worker_threads[-1].join()
                self.worker_threads.pop(-1)

            while len(self.worker_threads) < self.target_threads:
                self.worker_threads.append(
                    threading.Thread(target=self.worker, args=(len(self.worker_threads),), daemon=True)
                )
                self.worker_threads[-1].start()

    def apply(self, iterator: Iterable[Tuple[str, str, Target]], *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs
        for script, shortname, target in iterator:
            self.queue.put((script, shortname, target), block=True)
        self.queue.join()

    def worker(self, ident: int) -> None:
        while True:
            if ident >= self.target_threads:
                return
            script, shortname, target = self.queue.get(block=True)
            with self.lock:
                self.live_tasks += 1
            try:
                kwargs: Dict[str, Any] = self.kwargs | {'script_shortname': shortname}
                with self.rlock.gen_rlock():
                    shoot(script, target, *self.args, **kwargs)
            except:
                logger.exception("Failed to shoot")
            finally:
                self.queue.task_done()
                with self.lock:
                    self.live_tasks -= 1

def get_script_service_name_and_hash(script: str, service_hint: Optional[str] = None) -> Tuple[Optional[str], Optional[str]]:
    with open(script, 'rb') as fp:
        script_bytes = fp.read()
    match = SERVICE_NAME_RE.search(script_bytes)
    h = hashlib.md5()
    h.update(script_bytes)
    hd = h.hexdigest()
    if b'x-shooter-ignore' in script_bytes:
        return None, hd
    if match:
        service = match.group(1).decode()
        return service, hd
    if service_hint is not None:
        return service_hint, hd
    return None, None

LIVE_PROCESSES: Set[subprocess.Popen[bytes]] = set()

def kill_live_processes() -> None:
    for proc in list(LIVE_PROCESSES):
        proc.kill()

def shoot(
    script: str,
    target: Target,
    timeout: Optional[int],
    logdir: Optional[str],
    flag_regex: re.Pattern[bytes],
    use_stdout: bool,
    mem_limit: float,
    script_shortname: Optional[str] = None,
    corpus_hash: Optional[str] = None,
) -> None:
    if timeout is None:
        timeout = 999999999
    start_time = datetime.datetime.now()
    deadline = time.time() + timeout
    cmd = [os.path.join('.', os.path.basename(script))]
    env = dict(os.environ)
    for name in ('team', 'team_name', 'teamname'):
        env[name] = env[name.upper()] = target.team.name
    for name in ('host', 'victim', 'enemy'):
        env[name] = env[name.upper()] = target.team.hostname
    for name in ('port',):
        env[name] = env[name.upper()] = str(target.service.port)
    if target.flag_id is not None:
        for name in ('flagid', 'flag_id', 'fid', 'id'):
            env[name] = env[name.upper()] = target.flag_id

    if not os.access(script, os.X_OK):
        if script.endswith('.py'):
            cmd.insert(0, 'python3')

    with open(script, 'rb') as fp:
        script_bytes = fp.read()
    h = hashlib.md5()
    h.update(script_bytes)
    script_hash = h.hexdigest()

    def preexec_limits() -> None:
        size_bytes = int(mem_limit * 1024 * 1024 * 1024)
        #resource.setrlimit(resource.RLIMIT_STACK, (size_bytes, size_bytes))
        resource.setrlimit(resource.RLIMIT_DATA, (size_bytes, size_bytes))
        os.setpgrp()

    proc = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=os.path.dirname(os.path.realpath(script)),
        preexec_fn=preexec_limits,
        env=env,
    )
    assert proc.stdout is not None
    LIVE_PROCESSES.add(proc)
    head_buf: List[bytes] = []
    tail_buf: List[bytes] = []
    buf_full = False
    BUF_SIZE = 20
    LINE_SIZE = 1024*4
    seen_flags: List[bytes] = []

    hit_timeout = False
    while True:
        r, _, _ = select.select([proc.stdout], [], [], max(0., deadline - time.time()))
        if not r:
            hit_timeout = True
            if use_stdout:
                print('TIMEOUT')
            break
        line = proc.stdout.readline(LINE_SIZE)
        if not line:
            break
        if use_stdout:
            sys.stdout.buffer.write(line)
            sys.stdout.buffer.flush()
        if len(head_buf) < BUF_SIZE:
            head_buf.append(line)
        else:
            tail_buf.append(line)
            if len(tail_buf) >= BUF_SIZE:
                tail_buf.pop(0)
                buf_full = True

        for flag in flag_regex.finditer(line):
            seen_flags.append(flag.group(0))
            while True:
                try:
                    SUBMISSION_QUEUE.put(Submission(flag=flag.group(0), target=target, script=script))
                except queue.Full:
                    logger.warning("Submission queue is full")
                    time.sleep(5)
                else:
                    break

    try:
        code = proc.wait(deadline - time.time())
        try:
            os.killpg(proc.pid, signal.SIGKILL)
        except ProcessLookupError:
            pass
        status = ScriptResult.SUCCESS if code == 0 else ScriptResult.FAILURE
    except subprocess.TimeoutExpired:
        try:
            os.killpg(proc.pid, signal.SIGKILL)
        except ProcessLookupError:
            pass
        code = proc.wait()
        status = ScriptResult.TIMEOUT
    LIVE_PROCESSES.remove(proc)

    metrics = ScriptMetrics(
            script=script_shortname or script,
            script_hash=script_hash,
            corpus_hash=corpus_hash,
            target=target,
            start_time=start_time,
            duration=datetime.datetime.now() - start_time,
            status=status,
            flags_seen=len(seen_flags),
            flags= tuple(f.decode('latin-1') for f in seen_flags),
            host=platform.node(),
            exit_code=code,
    )
    METRICS_QUEUE.put(metrics)

    if logdir is not None:
        log_filename = os.path.join(logdir, script_shortname or os.path.basename(script), start_time.isoformat() + '-' + target.team.name + '.log')
        pathlib.Path(log_filename).parent.mkdir(parents=True, exist_ok=True)
        with open(log_filename, 'wb') as fp2:
            fp2.write(metrics.to_json().encode() + b'\n')
            fp2.writelines(head_buf)
            if buf_full:
                fp2.write(b'...\n')
            fp2.writelines(tail_buf)
            if hit_timeout:
                fp2.write(b'TIMEOUT\n')


SUBMISSIONS_DONE: bool = False
SUBMISSION_QUEUE: "queue.Queue[Submission]" = queue.Queue(maxsize=10000)
METRICS_QUEUE: "queue.Queue[ScriptMetrics]" = queue.Queue(maxsize=10000)
NOTIFIED_SOCKS: List[nclib.Netcat] = []
FLAG_MAPPING: Dict[bytes, Submission] = {}

def metrics_routine(server: str, debounce: Union[int, float]) -> None:
    while not SUBMISSIONS_DONE:
        buffer = set()
        deadline = time.time() + debounce
        while True:
            try:
                metric = METRICS_QUEUE.get(block=True, timeout=deadline - time.time())
            except (queue.Empty, ValueError):
                break
            else:
                buffer.add(metric)

        if buffer:
            try:
                sock = nclib.Netcat(server)
            except:
                logger.exception('Could not connect to server for metrics submission')
                continue
            try:
                sock.sendln(b'pubmetrics')
                sock.send(b''.join(s.to_json().encode() + b'\n' for s in buffer))
                sock.close()
            except:
                logger.exception('Failed to submit metrics')

def submission_routine(server: str, debounce: Union[int, float]) -> None:
    buffer = set()
    while not SUBMISSIONS_DONE:
        deadline = time.time() + debounce
        while True:
            try:
                submission = SUBMISSION_QUEUE.get(block=True, timeout=deadline - time.time())
            except (queue.Empty, ValueError):  # ValueError = negative timeout
                break
            else:
                buffer.add(submission)
                FLAG_MAPPING[submission.flag.strip()] = submission

        if buffer:
            try:
                sock = nclib.Netcat(server)
            except:
                logger.exception('Could not connect to server for flag submission')
                continue
            try:
                logger.info("Submitting %d flags", len(buffer))
                sock.sendln(b'submit')
                sock.send(b''.join(s.to_json().encode() + b'\n' for s in buffer))
                sock.shutdown_wr()
            except:
                logger.exception('Failed to submit flags')
            else:
                buffer.clear()
            NOTIFIED_SOCKS.append(sock)

def notification_routine() -> None:
    while not SUBMISSIONS_DONE or NOTIFIED_SOCKS:
        r, _, _ = nclib.select(NOTIFIED_SOCKS, [], [], timeout=1)
        for sock in r:
            line = sock.recvln()
            if not line:
                NOTIFIED_SOCKS.remove(sock)
            else:
                submission = FLAG_MAPPING.get(line.strip(), None)
                if submission is None:
                    logger.warning("Got points on unknown flag?")
                else:
                    logger.info("Got points on %s:%s", submission.target.team.name, submission.target.service.name)

def submit_routine(server: str, target: Target, script: str) -> None:
    flag_regex = get_flag_regex(server)
    for line in sys.stdin.buffer:
        for flag in flag_regex.finditer(line):
            SUBMISSION_QUEUE.put(Submission(flag=flag.group(0), target=target, script=script))

def sync_routine(command: str, interval: int, lock: rwlock.Lockable) -> None:
    while not SUBMISSIONS_DONE:
        with lock:
            subprocess.run(command, shell=True, stdin=subprocess.DEVNULL)
        time.sleep(interval)


def main() -> None:
    global SUBMISSIONS_DONE
    setup_logging()
    ctrl_c_times = 0
    def force_exit(*args: Any) -> None:
        nonlocal ctrl_c_times
        global SUBMISSIONS_DONE
        if ctrl_c_times == 0:
            sys.stderr.write("CTRL-C again to force exit\n")
            sys.stderr.flush()
            ctrl_c_times += 1
        else:
            sys.stderr.write("Cancelling all tasks and exiting\n")
            sys.stderr.flush()
            SUBMISSIONS_DONE = True
            sys.exit()

    def on_shutdown(*args: Any) -> None:
        global SUBMISSIONS_DONE
        SUBMISSIONS_DONE = True
        sys.exit()

    args = parser.parse_args(sys.argv[1:], namespace=ShooterArgs())
    if args.status_html:
        with open(args.status_html, 'w', encoding='utf-8') as fp:
            fp.write(statuspage(args.server))
        return
    if args.status_server:
        host, port_s = args.status_server.split(':')
        port = int(port_s)
        server = StatusServer((host, port), StatusHandler, args.server)
        print(f'serving on http://{args.status_server}/')
        server.serve_forever()
        return

    signal.signal(signal.SIGINT, force_exit)
    signal.signal(signal.SIGTERM, on_shutdown)

    target_mode = parse_target_mode(args)
    debounce = 5 if isinstance(target_mode, Forever) or args.submit else 1

    submission_thread = threading.Thread(target=submission_routine, args=(args.server, debounce))
    notification_thread = threading.Thread(target=notification_routine)

    if args.submit:
        if args.script is None or args.host is None or args.port is None:
            sys.stderr.write("Error: must specify --script, --host, and --port so we know the providence of flags")
            sys.stderr.flush()
            sys.exit(1)

        submission_thread.start()
        notification_thread.start()

        assert isinstance(target_mode, Single)
        target = Target(
            team=Team(name=target_mode.host, hostname=target_mode.host),
            service=Service(name=str(target_mode.port), port=target_mode.port),
            flag_id=target_mode.flag_id,
        )
        submit_routine(args.server, target, args.script)
        SUBMISSIONS_DONE = True
    else:
        submission_thread.start()
        notification_thread.start()

        mgr: Iterable[Tuple[str, str, Target]]
        if args.corpus is not None:
            if args.script is not None:
                sys.stderr.write("Error: only specify one of --corpus or --script\n")
                sys.stderr.flush()
                sys.exit(1)
            mgr = CorpusManager(args.corpus, args.server, args.batch, target_mode, args.corpus_hints)
        elif args.script is not None:
            mgr = ScriptManager(args.script, args.server, args.batch, target_mode, args.script, toplevel=True)
        else:
            sys.stderr.write("Error: must specify one of --corpus or --script\n")
            sys.stderr.flush()
            sys.exit(1)

        pool: Pool
        rlock = rwlock.RWLockWrite()
        if args.adaptive_procs:
            pool = AdaptivePool(rlock)
        elif args.procs:
            pool = AsyncPool(rlock, args.procs)
        else:
            pool = SynchronousPool(rlock)

        metrics_thread = threading.Thread(target=metrics_routine, args=(args.server, debounce))
        metrics_thread.start()

        sync_thread = None
        if args.sync_command is not None:
            sync_thread = threading.Thread(target=sync_routine, args=(args.sync_command, args.sync_interval, rlock.gen_wlock()))
            sync_thread.start()

        pool.apply(
            mgr,
            timeout=args.timeout,
            logdir=args.logdir,
            flag_regex=get_flag_regex(args.server),
            use_stdout=not args.no_stdout,
            mem_limit=args.mem_limit,
        )

        SUBMISSIONS_DONE = True
        metrics_thread.join()
        if sync_thread is not None:
            sync_thread.join()

    submission_thread.join()
    notification_thread.join()

if __name__ == '__main__':
    main()
